# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The acls parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""
import re

from ansible.module_utils._text import to_text
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def remarks_with_sequence(remarks_data):
    cmd = "remark"
    if remarks_data.get("remarks"):
        cmd += " " + remarks_data.get("remarks")
    if remarks_data.get("sequence"):
        if remarks_data.get("afi") == "ipv6":
            cmd = "sequence " + to_text(remarks_data.get("sequence")) + " " + cmd
        else:
            cmd = to_text(remarks_data.get("sequence")) + " " + cmd
    return cmd


def _tmplt_access_list_entries(aces):
    def source_destination_common_config(config_data, command, attr):
        source_host = config_data.get("source").get("host") or config_data.get("source").get(
            "address",
        )
        source_port_protocol = config_data.get("source").get("port_protocol")
        any_source_used = False
        if config_data[attr].get("any") and attr == "destination":
            command += " any"
            source_object = config_data.get("source").get("object_group")
            source_any = config_data.get("source").get("any")
            if (
                not source_host
                and not source_any
                and not source_object
                and not source_port_protocol
            ):
                command += " any"

        if config_data[attr].get("address"):
            command += " {address}".format(**config_data[attr])
            if config_data[attr].get("wildcard_bits"):
                command += " {wildcard_bits}".format(**config_data[attr])
        elif config_data[attr].get("host"):
            command += " host {host}".format(**config_data[attr])
        if config_data[attr].get("object_group"):
            command += " object-group {object_group}".format(**config_data[attr])
        if config_data[attr].get("port_protocol"):
            if (
                source_port_protocol
                and not source_host
                and config_data.get("source").get("any")
                and attr == "source"
            ):
                command += " any"
                any_source_used = True
            if config_data[attr].get("port_protocol").get("range"):
                command += " range {0} {1}".format(
                    config_data[attr]["port_protocol"]["range"].get("start"),
                    config_data[attr]["port_protocol"]["range"].get("end"),
                )
            else:
                port_proto_type = list(config_data[attr]["port_protocol"].keys())[0]
                command += " {0} {1}".format(
                    port_proto_type,
                    config_data[attr]["port_protocol"][port_proto_type],
                )

        if config_data[attr].get("any") and attr == "source" and not any_source_used:
            command += " any"
        return command

    command = ""
    proto_option = None
    if aces:
        if aces.get("sequence") and aces.get("afi") == "ipv4":
            command += "{sequence}".format(**aces)
        if aces.get("grant") and aces.get("sequence") and aces.get("afi") == "ipv4":
            command += " {grant}".format(**aces)
        elif aces.get("grant") and aces.get("sequence") and aces.get("afi") == "ipv6":
            command += "{grant}".format(**aces)
        elif aces.get("grant"):
            command += "{grant}".format(**aces)
        if aces.get("protocol_options"):
            if "protocol_number" in aces["protocol_options"]:
                command += " {protocol_number}".format(**aces["protocol_options"])
            else:
                command += " {0}".format(list(aces["protocol_options"])[0])
                proto_option = aces["protocol_options"].get(
                    list(aces["protocol_options"])[0],
                )
        elif aces.get("protocol"):
            command += " {protocol}".format(**aces)
        if aces.get("source"):
            command = source_destination_common_config(aces, command, "source")
        if aces.get("destination"):
            command = source_destination_common_config(aces, command, "destination")
        if isinstance(proto_option, dict):
            for flag, enabled in proto_option.items():
                if enabled:
                    command += " {}".format(flag.replace("_", "-"))
        if aces.get("dscp"):
            command += " dscp {dscp}".format(**aces)
        if aces.get("sequence") and aces.get("afi") == "ipv6":
            command += " sequence {sequence}".format(**aces)
        if aces.get("enable_fragments"):
            command += " fragments"
        if aces.get("log"):
            command += " log"
            if aces["log"].get("user_cookie"):
                command += " {user_cookie}".format(**aces["log"])
        if aces.get("log_input"):
            command += " log-input"
            if aces["log_input"].get("user_cookie"):
                command += " {user_cookie}".format(**aces["log_input"])
        if aces.get("option"):
            option_val = list(aces.get("option").keys())[0]
            command += " option {0}".format(option_val)
        if aces.get("precedence"):
            command += " precedence {precedence}".format(**aces)
        if aces.get("time_range"):
            command += " time-range {time_range}".format(**aces)
        if aces.get("tos"):
            command += " tos"
            if aces["tos"].get("service_value"):
                command += " {service_value}".format(**aces["tos"])
            elif aces["tos"].get("max_reliability"):
                command += " max-reliability"
            elif aces["tos"].get("max_throughput"):
                command += " max-throughput"
            elif aces["tos"].get("min_delay"):
                command += " min-delay"
            elif aces["tos"].get("min_monetary_cost"):
                command += " min-monetary-cost"
            elif aces["tos"].get("normal"):
                command += " normal"
        if aces.get("ttl"):
            command += " ttl {0}".format(list(aces["ttl"])[0])
            proto_option = aces["ttl"].get(list(aces["ttl"])[0])
            command += " {0}".format(proto_option)
    return command


class AclsTemplate(NetworkTemplate):
    def __init__(self, lines=None):
        super(AclsTemplate, self).__init__(lines=lines, tmplt=self)

    PARSERS = [
        {
            "name": "only_acls_name",
            "getval": re.compile(
                r"""^(?P<acl_type>Standard|Extended|Reflexive)*
                    \s*(?P<afi>IP|IPv6)*
                    \s*access*
                    \s*list*
                    \s*(?P<acl_name>.+)*
                    $""",
                re.VERBOSE,
            ),
            "setval": "",
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "acl_type": "{{ acl_type.lower() if acl_type is defined }}",
                        "afi": "{{ 'ipv4' if afi == 'IP' else 'ipv6' }}",
                    },
                },
            },
        },
        {
            "name": "acls_name",
            "getval": re.compile(
                r"""^(?P<afi>ip|ipv6|mac)
                    (\s(access-list))
                    (\s(?P<acl_type>standard|extended|reflexive))?
                    (\s(?P<acl_name>\S+))
                    $""",
                re.VERBOSE,
            ),
            "setval": "name",
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "acl_type": "{{ acl_type.lower() if acl_type is defined }}",
                        "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "remarks",
            "getval": re.compile(
                r"""((?P<order>^\d+))
                    (\s*(?P<sequence>\d+))
                    (\sremark\s(?P<remarks>.+))
                    $""",
                re.VERBOSE,
            ),
            "setval": remarks_with_sequence,
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "aces": [
                            {
                                "the_remark": "'{{ remarks }}'",
                                "order": "{{ order }}",
                                "is_remark_for": "{{ sequence }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "remarks_no_data",
            "getval": re.compile(
                r"""(?P<order>^\d+)\s*remark\s(?P<remarks>.+)$""",
                re.VERBOSE,
            ),
            "setval": remarks_with_sequence,
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "aces": [
                            {
                                "the_remark": "'{{ remarks }}'",
                                "order": "{{ order }}",
                                "is_remark_for": "remark",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "remarks_ipv6",
            "getval": re.compile(
                r"""\s*(sequence\s(?P<sequence>\d+))
                    (\sremark\s(?P<remarks>.+))
                    $""",
                re.VERBOSE,
            ),
            "setval": remarks_with_sequence,
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "aces": [
                            {
                                "sequence": "{{ sequence }}",
                                "remarks": ["'{{ remarks }}'"],
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "aces_ipv4_standard",
            "getval": re.compile(
                r"""^\s*((?P<sequence>\d+))?
                        (\s(?P<grant>deny|permit))
                        (\s+(?P<address>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))?
                        (\s(?P<wildcard>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))?
                        (\s*(?P<any>any))?
                        (\s(?P<log>log))?
                    $""",
                re.VERBOSE,
            ),
            "compval": "aces",
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "aces": [
                            {
                                "sequence": "{{ sequence }}",
                                "grant": "{{ grant }}",
                                "source": {
                                    "address": "{{ address }}",
                                    "wildcard_bits": "{{ wildcard }}",
                                    "any": "{{ not not any }}",
                                },
                                "log": {"set": "{{ not not log }}"},
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "aces_source_dest_any",
            "getval": re.compile(
                r"""
                ^\s*((?P<sequence>\d+))?
                (\s(?P<grant>deny|permit))
                ((\s*object-group\s(?P<source_obj_grp>\S+))|
                (\s*host\s(?P<source_host>\S+))|
                (\s*(?P<ipv6_source_address>\S+/\d+))|
                (\s*(?P<source_address>(\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3})\s\S+)))?
                (\s*(?P<source_any>any))
                (\s*(?P<dest_any>any))
                $""",
                re.VERBOSE,
            ),
            "compval": "aces",
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "aces": [
                            {
                                "sequence": "{{ sequence }}",
                                "grant": "{{ grant }}",
                                "source": {
                                    "address": "{{ source_address }}",
                                    "ipv6_address": "{{ ipv6_source_address }}",
                                    "any": "{{ not not source_any }}",
                                    "host": "{{ source_host }}",
                                    "object_group": "{{ source_obj_grp }}",
                                },
                                "destination": {"any": "{{ not not dest_any }}"},
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "aces",
            "getval": re.compile(
                r"""^\s*((?P<sequence>\d+))?
                        (\s(?P<grant>deny|permit))
                        (\sevaluate\s(?P<evaluate>\S+))?
                        (\s(?P<protocol_num>\d+)\s)?
                        (\s*(?P<protocol>ahp|eigrp|esp|gre|icmp|igmp|ipinip|ipv6|ip|nos|ospf|pcp|pim|sctp|tcp|udp))?
                        (?:(\s*(?P<source_any>any))|
                        (\s*object-group\s(?P<source_obj_grp>\S+))|
                        (\s*host\s(?P<source_host>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)
                            {3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))|
                        (\s*(?P<source_address>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)
                            {3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))
                        (\s+(?P<source_wildcard>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)
                            {3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))?))
                        (\seq\s(?P<seq>(\S+|\d+)))?
                        (\sgt\s(?P<sgt>(\S+|\d+)))?
                        (\slt\s(?P<slt>(\S+|\d+)))?
                        (\sneq\s(?P<sneq>(\S+|\d+)))?
                        (\srange\s(?P<srange_start>\S+)\s(?P<srange_end>\S+))?
                        (?:(\s*(?P<dest_any>any))|
                        (\s*object-group\s(?P<dest_obj_grp>\S+))|
                        (\s*host\s(?P<dest_host>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)
                            {3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))|
                        (\s*(?P<dest_address>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)
                            {3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))
                        (\s+(?P<dest_wildcard>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)
                            {3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))?))
                        (\seq\s(?P<deq>(\S+|\d+)))?
                        (\sgt\s(?P<dgt>(\S+|\d+)))?
                        (\slt\s(?P<dlt>(\S+|\d+)))?
                        (\sneq\s(?P<dneq>(\S+|\d+)))?
                        (\srange\s(?P<drange_start>\S+)\s(?P<drange_end>\S+))?
                        (\s+(?P<icmp_igmp_protocol>
                        administratively-prohibited|alternate-address|conversion-error|dod-host-prohibited|dod-net-prohibited|
                        echo-reply|echo|general-parameter-problem|host-isolated|host-precedence-unreachable|host-redirect|
                        host-tos-redirect|host-tos-unreachable|host-unknown|host-unreachable|information-reply|information-request|
                        mask-reply|mask-request|mobile-redirect|net-redirect|net-tos-redirect|net-tos-unreachable|net-unreachable|
                        network-unknown|no-room-for-option|option-missing|packet-too-big|parameter-problem|port-unreachable|
                        precedence-unreachable|protocol-unreachable|reassembly-timeout|redirect|router-advertisement|
                        router-solicitation|source-quench|source-route-failed|time-exceeded|timestamp-reply|timestamp-request|
                        traceroute|ttl-exceeded|unreachable|dvmrp|host-query|mtrace-resp|mtrace-route|pim|trace|
                        v1host-report|v2host-report|v2leave-group|v3host-report|\d+))?
                        (\s+(?P<tcp_flags>(?:ack|fin|psh|rst|syn|urg|established)(?:\s+(?:ack|fin|psh|rst|syn|urg|established))*))?
                        (\sdscp\s(?P<dscp>\S+))?
                        (\s(?P<enable_fragments>fragments))?
                        (\slog-input\s\(tag\s=\s(?P<log_input>\S+)\))?
                        (\s(?P<log_input_only>log-input))?
                        (\slog\s\(tag\s=\s(?P<log>\S+)\))?
                        (\s(?P<log_only>log))?
                        (\soption\s(?P<option>\S+|\d+))?
                        (\sprecedence\s(?P<precedence>\S+))?
                        (\stime-range\s(?P<time_range>\S+))?
                        (\stos\s(?P<tos>\S+|\d+))?
                        (\sttl\seq\s(?P<ttl_eq>\d+))?
                        (\sttl\sgt\s(?P<ttl_gt>\d+))?
                        (\sttl\slt\s(?P<ttl_lt>\d+))?
                        (\sttl\sneg\s(?P<ttl_neg>\d+))?
                    """,
                re.VERBOSE,
            ),
            "setval": _tmplt_access_list_entries,
            "compval": "aces",
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "aces": [
                            {
                                "sequence": "{{ sequence }}",
                                "grant": "{{ grant }}",
                                "evaluate": "{{ evaluate }}",
                                "protocol": "{{ protocol }}",
                                "protocol_number": "{{ protocol_num }}",
                                "protocol_options": "{{ tcp_flags if tcp_flags is defined"
                                " else (icmp_igmp_protocol if protocol is defined"
                                " and protocol in ['icmp', 'igmp'] else undefined) }}",
                                "source": {
                                    "address": "{{ source_address }}",
                                    "wildcard_bits": "{{ source_wildcard }}",
                                    "any": "{{ not not source_any }}",
                                    "host": "{{ source_host }}",
                                    "object_group": "{{ source_obj_grp }}",
                                    "port_protocol": {
                                        "eq": "{{ seq }}",
                                        "gt": "{{ sgt }}",
                                        "lt": "{{ slt }}",
                                        "neq": "{{ sneq }}",
                                        "range": {
                                            "start": "{{ srange_start if srange_start is defined else None }}",
                                            "end": "{{ srange_end if srange_end is defined else None }}",
                                        },
                                    },
                                },
                                "destination": {
                                    "address": "{{ dest_address }}",
                                    "wildcard_bits": "{{ dest_wildcard }}",
                                    "any": "{{ not not dest_any }}",
                                    "host": "{{ dest_host }}",
                                    "object_group": "{{ dest_obj_grp }}",
                                    "port_protocol": {
                                        "eq": "{{ deq }}",
                                        "gt": "{{ dgt }}",
                                        "lt": "{{ dlt }}",
                                        "neq": "{{ dneq }}",
                                        "range": {
                                            "start": "{{ drange_start if drange_start is defined else None }}",
                                            "end": "{{ drange_end if drange_end is defined else None }}",
                                        },
                                    },
                                },
                                "dscp": "{{ dscp }}",
                                "enable_fragments": "{{ True if enable_fragments is defined else None }}",
                                "log": {
                                    "set": "{{ True if log_only is defined or log is defined }}",
                                    "user_cookie": "{{ log if log is defined else None }}",
                                },
                                "log_input": {
                                    "set": "{{ True if log_input_only is defined or log_input is defined }}",
                                    "user_cookie": "{{ log_input if log_input is defined else None }}",
                                },
                                "option": {
                                    "{{ option if option is defined else None }}": "{{ True if option is defined else None }}",
                                },
                                "precedence": "{{ precedence }}",
                                "time_range": "{{ time_range }}",
                                "tos": {
                                    "max_reliability": "{{ True if tos is defined and 'max-reliability' in tos else '' }}",
                                    "max_throughput": "{{ True if tos is defined and 'max-throughput' in tos else '' }}",
                                    "min_delay": "{{ True if tos is defined and 'min-delay' in tos else '' }}",
                                    "min_monetary_cost": "{{ True if tos is defined and 'min-monetary-cost' in tos else '' }}",
                                    "normal": "{{ True if tos is defined and 'normal' in tos else '' }}",
                                    "service_value": "{{ tos if tos is defined else None }}",
                                },
                                "ttl": {
                                    "eq": "{{ ttl_eq }}",
                                    "gt": "{{ ttl_gt }}",
                                    "lt": "{{ ttl_lt }}",
                                    "neq": "{{ ttl_neq }}",
                                },
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "aces_ipv6",
            "getval": re.compile(
                r"""^\s*(sequence\s+(?P<sequence_ipv6>\d+))?
                        (\s(?P<grant>deny|permit))
                        (\sevaluate\s(?P<evaluate>\S+))?
                        (\s(?P<protocol_num>\d+)\s)?
                        (\s*(?P<protocol>ahp|eigrp|esp|gre|icmp|igmp|ipinip|ipv6|ip|nos|ospf|pcp|pim|sctp|tcp|udp))?
                        (\s*(?P<source_any>any)|
                        \s*object-group\s(?P<source_obj_grp>\S+)|
                        \s*host\s(?P<source_host>\S+)|
                        \s*(?P<ipv6_source_address>\S+)
                        (?:\s+(?P<source_wildcard_v6>(?![^/\s]+\/)[^/\s]*:[^/\s]*))?)
                        (\seq\s(?P<seq>(\S+|\d+)))?
                        (\sgt\s(?P<sgt>(\S+|\d+)))?
                        (\slt\s(?P<slt>(\S+|\d+)))?
                        (\sneq\s(?P<sneq>(\S+|\d+)))?
                        (\srange\s(?P<srange_start>\S+)\s(?P<srange_end>\S+))?
                        (\s*(?P<dest_any>any)|
                        \s*object-group\s(?P<dest_obj_grp>\S+)|
                        \s*host\s(?P<dest_host>\S+)|
                        \s*(?P<ipv6_dest_address>\S+)
                        (?:\s+(?P<dest_wildcard_v6>(?![^/\s]+\/)[^/\s]*:[^/\s]*))?)
                        (\seq\s(?P<deq>(\S+|\d+)))?
                        (\sgt\s(?P<dgt>(\S+|\d+)))?
                        (\slt\s(?P<dlt>(\S+|\d+)))?
                        (\sneq\s(?P<dneq>(\S+|\d+)))?
                        (\srange\s(?P<drange_start>\S+)\s(?P<drange_end>\S+))?
                        (\s+(?P<icmp_igmp_protocol>
                        administratively-prohibited|alternate-address|conversion-error|dod-host-prohibited|dod-net-prohibited|
                        echo-reply|echo|general-parameter-problem|host-isolated|host-precedence-unreachable|host-redirect|
                        host-tos-redirect|host-tos-unreachable|host-unknown|host-unreachable|information-reply|information-request|
                        mask-reply|mask-request|mobile-redirect|net-redirect|net-tos-redirect|net-tos-unreachable|net-unreachable|
                        network-unknown|no-room-for-option|option-missing|packet-too-big|parameter-problem|port-unreachable|
                        precedence-unreachable|protocol-unreachable|reassembly-timeout|redirect|router-advertisement|
                        router-solicitation|source-quench|source-route-failed|time-exceeded|timestamp-reply|timestamp-request|
                        traceroute|ttl-exceeded|unreachable|dvmrp|host-query|mtrace-resp|mtrace-route|pim|trace|
                        v1host-report|v2host-report|v2leave-group|v3host-report|\d+))?
                        (\s+(?P<tcp_flags>(?:ack|fin|psh|rst|syn|urg|established)(?:\s+(?:ack|fin|psh|rst|syn|urg|established))*))?
                        (\sdscp\s(?P<dscp>\S+))?
                        (\s(?P<enable_fragments>fragments))?
                        (\slog-input\s\(tag\s=\s(?P<log_input>\S+)\))?
                        (\s(?P<log_input_only>log-input))?
                        (\slog\s\(tag\s=\s(?P<log>\S+)\))?
                        (\s(?P<log_only>log))?
                        (\soption\s(?P<option>\S+|\d+))?
                        (\sprecedence\s(?P<precedence>\S+))?
                        (\stime-range\s(?P<time_range>\S+))?
                        (\stos\s(?P<tos>\S+|\d+))?
                        (\sttl\seq\s(?P<ttl_eq>\d+))?
                        (\sttl\sgt\s(?P<ttl_gt>\d+))?
                        (\sttl\slt\s(?P<ttl_lt>\d+))?
                        (\sttl\sneg\s(?P<ttl_neg>\d+))?
                    """,
                re.VERBOSE,
            ),
            "setval": _tmplt_access_list_entries,
            "compval": "aces",
            "result": {
                "acls": {
                    "{{ acl_name|d() }}": {
                        "name": "{{ acl_name }}",
                        "aces": [
                            {
                                "sequence": "{{ sequence_ipv6 }}",
                                "grant": "{{ grant }}",
                                "evaluate": "{{ evaluate }}",
                                "protocol": "{{ protocol }}",
                                "protocol_number": "{{ protocol_num }}",
                                "protocol_options": "{{ tcp_flags if tcp_flags is defined"
                                " else (icmp_igmp_protocol if protocol is defined"
                                " and protocol in ['icmp', 'igmp'] else undefined) }}",
                                "source": {
                                    "address": "{{ ipv6_source_address }}",
                                    "wildcard_bits": "{{ source_wildcard_v6 }}",
                                    "any": "{{ not not source_any }}",
                                    "host": "{{ source_host }}",
                                    "object_group": "{{ source_obj_grp }}",
                                    "port_protocol": {
                                        "eq": "{{ seq }}",
                                        "gt": "{{ sgt }}",
                                        "lt": "{{ slt }}",
                                        "neq": "{{ sneq }}",
                                        "range": {
                                            "start": "{{ srange_start if srange_start is defined else None }}",
                                            "end": "{{ srange_end if srange_end is defined else None }}",
                                        },
                                    },
                                },
                                "destination": {
                                    "address": "{{ ipv6_dest_address }}",
                                    "wildcard_bits": "{{ dest_wildcard_v6 }}",
                                    "any": "{{ not not dest_any }}",
                                    "host": "{{ dest_host }}",
                                    "object_group": "{{ dest_obj_grp }}",
                                    "port_protocol": {
                                        "eq": "{{ deq }}",
                                        "gt": "{{ dgt }}",
                                        "lt": "{{ dlt }}",
                                        "neq": "{{ dneq }}",
                                        "range": {
                                            "start": "{{ drange_start if drange_start is defined else None }}",
                                            "end": "{{ drange_end if drange_end is defined else None }}",
                                        },
                                    },
                                },
                                "dscp": "{{ dscp }}",
                                "enable_fragments": "{{ True if enable_fragments is defined else None }}",
                                "log": {
                                    "set": "{{ True if log_only is defined or log is defined }}",
                                    "user_cookie": "{{ log if log is defined else None }}",
                                },
                                "log_input": {
                                    "set": "{{ True if log_input_only is defined or log_input is defined }}",
                                    "user_cookie": "{{ log_input if log_input is defined else None }}",
                                },
                                "option": {
                                    "{{ option if option is defined else None }}": "{{ True if option is defined else None }}",
                                },
                                "precedence": "{{ precedence }}",
                                "time_range": "{{ time_range }}",
                                "tos": {
                                    "max_reliability": "{{ True if tos is defined and 'max-reliability' in tos else '' }}",
                                    "max_throughput": "{{ True if tos is defined and 'max-throughput' in tos else '' }}",
                                    "min_delay": "{{ True if tos is defined and 'min-delay' in tos else '' }}",
                                    "min_monetary_cost": "{{ True if tos is defined and 'min-monetary-cost' in tos else '' }}",
                                    "normal": "{{ True if tos is defined and 'normal' in tos else '' }}",
                                    "service_value": "{{ tos if tos is defined else None }}",
                                },
                                "ttl": {
                                    "eq": "{{ ttl_eq }}",
                                    "gt": "{{ ttl_gt }}",
                                    "lt": "{{ ttl_lt }}",
                                    "neq": "{{ ttl_neq }}",
                                },
                            },
                        ],
                    },
                },
            },
        },
    ]



#!/usr/bin/python
#
# Copyright (c) 2019 Yunge Zhu, <yungez@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_monitorlogprofile
version_added: "0.0.1"
short_description: Manage Azure Monitor log profile
description:
    - Create, update and delete instance of Azure Monitor log profile.

options:
    name:
        description:
            - Unique name of the log profile to create or update.
        required: True
        type: str
    location:
        description:
            - Resource location.
        type: str
    locations:
        description:
            - List of regions for which Activity Log events should be stored.
        type: list
        elements: str
    categories:
        description:
            - List of categories of logs. These categories are created as is convenient to  user. Some Values are C(Write), C(Delete) and/or C(Action).
        type: list
        elements: str
    retention_policy:
        description:
            - Retention policy for events in the log.
        type: dict
        suboptions:
            enabled:
                description:
                    - Whether the retention policy is enabled.
                type: bool
            days:
                description:
                    - The number of days for the retention. A value of 0 will retain the events indefinitely.
                type: int
    service_bus_rule_id:
        description:
            - The service bus rule  ID of the service bus namespace in which you would like to have Event Hubs created for streaming in the Activity Log.
            - Format like {service_bus_resource_id}/authorizationrules{key_name}.
        type: str
    storage_account:
        description:
            - The storage account to which send the Activity Log.
            - It could be a resource ID.
            - It could be a dict containing I(resource_grorup) and I(name).
        type: raw
    state:
        description:
            - Assert the state of the log profile.
            - Use C(present) to create or update a log profile and C(absent) to delete it.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Yunge Zhu(@yungezz)

'''

EXAMPLES = '''
- name: Create a log profile
  azure_rm_monitorlogprofile:
    name: myProfile
    location: eastus
    locations:
      - eastus
      - westus
    categories:
      - Write
      - Action
    retention_policy:
      enabled: false
      days: 1
    storage_account:
      resource_group: myResourceGroup
      name: myStorageAccount
  register: output

- name: Delete a log profile
  azure_rm_monitorlogprofile:
    name: myProfile
    state: absent
'''

RETURN = '''
id:
    description:
        - ID of the log profile.
    returned: always
    type: str
    sample: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/providers/microsoft.insights/logprofiles/myProfile

'''


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase, format_resource_id
    from azure.core.exceptions import HttpResponseError
    from azure.core.polling import LROPoller
    from azure.mgmt.core.tools import is_valid_resource_id
    from azure.mgmt.monitor.models import (RetentionPolicy, LogProfileResource)
except ImportError:
    # This is handled in azure_rm_common
    pass


retention_policy_spec = dict(
    enabled=dict(type='bool'),
    days=dict(type='int')
)


def logprofile_to_dict(profile):
    return dict(
        id=profile.id,
        name=profile.name,
        location=profile.location,
        locations=profile.locations,
        categories=profile.categories,
        storage_account=profile.storage_account_id,
        service_bus_rule_id=profile.service_bus_rule_id,
        retention_policy=dict(
            enabled=profile.retention_policy.enabled,
            days=profile.retention_policy.days
        ),
        tags=profile.tags if profile.tags else None
    )


class Actions:
    NoAction, CreateOrUpdate, Delete = range(3)


class AzureRMMonitorLogprofile(AzureRMModuleBase):
    """Configuration class for an Azure RM Monitor log profile"""

    def __init__(self):
        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=True
            ),
            location=dict(
                type='str'
            ),
            locations=dict(
                type='list',
                elements='str'
            ),
            categories=dict(
                type='list',
                elements='str'
            ),
            retention_policy=dict(
                type='dict',
                options=retention_policy_spec
            ),
            service_bus_rule_id=dict(
                type='str'
            ),
            storage_account=dict(
                type='raw'
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            )
        )

        self._client = None

        self.name = None
        self.location = None

        self.locations = None
        self.categories = None
        self.retention_policy = False
        self.service_bus_rule_id = None
        self.storage_account = None

        self.tags = None

        self.results = dict(
            changed=False,
            id=None
        )
        self.state = None

        super(AzureRMMonitorLogprofile, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                       supports_check_mode=True,
                                                       supports_tags=True)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        old_response = None
        response = None
        to_be_updated = False

        # get storage account id
        if self.storage_account:
            if isinstance(self.storage_account, dict):
                self.storage_account = format_resource_id(val=self.storage_account['name'],
                                                          subscription_id=self.storage_account.get('subscription') or self.subscription_id,
                                                          namespace='Microsoft.Storage',
                                                          types='storageAccounts',
                                                          resource_group=self.storage_account.get('resource_group'))
            elif not is_valid_resource_id(self.storage_account):
                self.fail("storage_account either be a resource id or a dict containing resource_group and name")

        # get existing log profile
        old_response = self.get_logprofile()

        if old_response:
            self.results['id'] = old_response['id']

        if self.state == 'present':
            # if profile not exists, create new
            if not old_response:
                self.log("Log profile instance doesn't exist")

                to_be_updated = True
                self.to_do = Actions.CreateOrUpdate

            else:
                # log profile exists already, do update
                self.log("Log profile instance already exists")

                update_tags, self.tags = self.update_tags(old_response.get('tags', None))

                if update_tags:
                    to_be_updated = True
                    self.to_do = Actions.CreateOrUpdate

                # check if update
                if self.check_update(old_response):
                    to_be_updated = True
                    self.to_do = Actions.CreateOrUpdate

        elif self.state == 'absent':
            if old_response:
                self.log("Delete log profile instance")
                self.results['id'] = old_response['id']
                to_be_updated = True
                self.to_do = Actions.Delete
            else:
                self.results['changed'] = False
                self.log("Log profile {0} not exists.".format(self.name))

        if to_be_updated:
            self.log('Need to Create/Update log profile')
            self.results['changed'] = True

            if self.check_mode:
                return self.results

            if self.to_do == Actions.CreateOrUpdate:
                response = self.create_or_update_logprofile()
                self.results['id'] = response['id']

            if self.to_do == Actions.Delete:
                self.delete_logprofile()
                self.log('Log profile instance deleted')

        return self.results

    def check_update(self, existing):
        if self.locations is not None and existing['locations'] != self.locations:
            self.log("locations diff: origin {0} / update {1}".format(existing['locations'], self.locations))
            return True
        if self.retention_policy is not None:
            if existing['retention_policy']['enabled'] != self.retention_policy['enabled']:
                self.log("retention_policy diff: origin {0} / update {1}".format(str(existing['sku']['name']), str(self.retention_policy['enabled'])))
                return True
            if existing['retention_policy']['days'] != self.retention_policy['days']:
                self.log("retention_policy diff: origin {0} / update {1}".format(existing['retention_policy']['days'], str(self.retention_policy['days'])))
                return True
        if self.storage_account is not None and existing['storage_account'] != self.storage_account:
            self.log("storage_account diff: origin {0} / update {1}".format(existing['storage_account'], self.storage_account))
            return True
        if self.service_bus_rule_id is not None and existing['service_bus_rule_id'] != self.service_bus_rule_id:
            self.log("service_bus_rule_id diff: origin {0} / update {1}".format(existing['service_bus_rule_id'], self.service_bus_rule_id))
            return True
        return False

    def create_or_update_logprofile(self):
        '''
        Creates or Update log profile.

        :return: deserialized log profile state dictionary
        '''
        self.log(
            "Creating log profile instance {0}".format(self.name))

        try:
            params = LogProfileResource(
                location=self.location,
                locations=self.locations,
                categories=self.categories,
                retention_policy=RetentionPolicy(days=self.retention_policy['days'],
                                                 enabled=self.retention_policy['enabled']) if self.retention_policy else None,
                storage_account_id=self.storage_account if self.storage_account else None,
                service_bus_rule_id=self.service_bus_rule_id if self.service_bus_rule_id else None,
                tags=self.tags
            )

            response = self.monitor_log_profiles_client.log_profiles.create_or_update(log_profile_name=self.name,
                                                                                      parameters=params)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)

        except HttpResponseError as exc:
            self.log('Error attempting to create/update log profile.')
            self.fail("Error creating/updating log profile: {0}".format(str(exc)))
        return logprofile_to_dict(response)

    def delete_logprofile(self):
        '''
        Deletes specified log profile.

        :return: True
        '''
        self.log("Deleting the log profile instance {0}".format(self.name))
        try:
            response = self.monitor_log_profiles_client.log_profiles.delete(log_profile_name=self.name)
        except HttpResponseError as e:
            self.log('Error attempting to delete the log profile.')
            self.fail(
                "Error deleting the log profile: {0}".format(str(e)))
        return True

    def get_logprofile(self):
        '''
        Gets the properties of the specified log profile.

        :return: log profile state dictionary
        '''
        self.log("Checking if the log profile {0} is present".format(self.name))

        response = None

        try:
            response = self.monitor_log_profiles_client.log_profiles.get(log_profile_name=self.name)

            self.log("Response : {0}".format(response))
            self.log("log profile : {0} found".format(response.name))
            return logprofile_to_dict(response)

        except HttpResponseError:
            self.log("Didn't find log profile {0}".format(self.name))

        return False


def main():
    """Main execution"""
    AzureRMMonitorLogprofile()


if __name__ == '__main__':
    main()



#
# (c) 2017 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = """
---
author: Egor Zaitsev (@heuels)
name: ciscosmb
short_description: Use ciscosmb cliconf to run command on Cisco SMB network devices
description:
  - This ciscosmb plugin provides low level abstraction apis for
    sending and receiving CLI commands from Cisco SMB network devices.
"""

import json

from ansible.plugins.cliconf import CliconfBase, enable_mode


class Cliconf(CliconfBase):

    def get_device_info(self):
        device_info = {}
        device_info["network_os"] = "ciscosmb"

        return device_info

    @enable_mode
    def get_config(self, source="running", flags=None, format=None):
        if source not in ("running", "startup"):
            raise ValueError("fetching configuration from %s is not supported" % source)

        if format:
            raise ValueError(
                "'format' value %s is not supported for get_config" % format
            )

        if flags:
            raise ValueError("'flags' value %s is not supported for get_config" % flags)

        if source == "running":
            cmd = "show running-config "
        else:
            cmd = "show startup-config "

        return self.send_command(cmd)

    def edit_config(self, command):
        return

    def get(
        self,
        command,
        prompt=None,
        answer=None,
        sendonly=False,
        newline=True,
        check_all=False,
    ):
        return self.send_command(
            command=command + "\n",
            prompt=prompt,
            answer=answer,
            sendonly=sendonly,
            newline=newline,
            check_all=check_all,
        )

    def get_capabilities(self):
        result = super().get_capabilities()
        return json.dumps(result)

    def get_serialization_profile(self, command=None):
        return "ciscosmb"  # This must match a profile defined elsewhere



#!/usr/bin/python
# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: nameserver_info
short_description: Look up nameservers for a DNS name
version_added: 2.6.0
description:
  - Retrieve all nameservers that are responsible for a DNS name.
extends_documentation_fragment:
  - community.dns._attributes
  - community.dns._attributes.info_module
  - community.dns._attributes.idempotent_not_modify_state
author:
  - Felix Fontein (@felixfontein)
options:
  name:
    description:
      - A list of DNS names whose nameservers to retrieve.
    required: true
    type: list
    elements: str
  resolve_addresses:
    description:
      - Whether to resolve the nameserver names to IP addresses.
    type: bool
    default: false
  query_retry:
    description:
      - Number of retries for DNS query timeouts.
    type: int
    default: 3
  query_timeout:
    description:
      - Timeout per DNS query in seconds.
    type: float
    default: 10
  always_ask_default_resolver:
    description:
      - When set to V(true) (default), will use the default resolver to find the authoritative nameservers of a subzone. See
        O(server) for how to configure the default resolver.
      - When set to V(false), will use the authoritative nameservers of the parent zone to find the authoritative nameservers
        of a subzone. This only makes sense when the nameservers were recently changed and have not yet propagated.
    type: bool
    default: true
  servfail_retries:
    description:
      - How often to retry on SERVFAIL errors.
    type: int
    default: 0
  server:
    description:
      - The DNS server(s) to use to look up the result. Must be a list of one or more IP addresses.
      - By default, the system's standard resolver is used.
    type: list
    elements: str
    version_added: 2.7.0
requirements:
  - dnspython >= 2.0.0
"""

EXAMPLES = r"""
- name: Retrieve name servers of two DNS names
  community.dns.nameserver_info:
    name:
      - www.example.com
      - example.org
  register: result

- name: Show nameservers for www.example.com
  ansible.builtin.debug:
    msg: '{{ result.results[0].nameserver }}'
"""

RETURN = r"""
results:
  description:
    - Information on the nameservers for every DNS name provided in O(name).
  returned: always
  type: list
  elements: dict
  contains:
    name:
      description:
        - The DNS name this entry is for.
      returned: always
      type: str
      sample: www.example.com
    nameservers:
      description:
        - A list of nameservers for this DNS name.
      returned: success
      type: list
      elements: str
      sample:
        - ns1.example.com
        - ns2.example.com
  sample:
    - name: www.example.com
      nameservers:
        - ns1.example.com
        - ns2.example.com
    - name: example.org
      nameservers:
        - ns1.example.org
        - ns2.example.org
        - ns3.example.org
"""

import typing as t

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.dns.plugins.module_utils._resolver import (
    ResolveDirectlyFromNameServers,
    assert_requirements_present,
    guarded_run,
)


def main() -> None:
    module = AnsibleModule(
        argument_spec={
            "name": {"required": True, "type": "list", "elements": "str"},
            "resolve_addresses": {"type": "bool", "default": False},
            "query_retry": {"type": "int", "default": 3},
            "query_timeout": {"type": "float", "default": 10},
            "always_ask_default_resolver": {"type": "bool", "default": True},
            "servfail_retries": {"type": "int", "default": 0},
            "server": {"type": "list", "elements": "str"},
        },
        supports_check_mode=True,
    )
    assert_requirements_present(module)

    names = module.params["name"]
    resolve_addresses = module.params["resolve_addresses"]

    resolver = ResolveDirectlyFromNameServers(
        timeout=module.params["query_timeout"],
        timeout_retries=module.params["query_retry"],
        servfail_retries=module.params["servfail_retries"],
        always_ask_default_resolver=module.params["always_ask_default_resolver"],
        server_addresses=module.params["server"],
    )
    results: list[dict[str, t.Any]] = [{"name": name} for name in names]

    def f():
        for index, name in enumerate(names):
            results[index]["nameservers"] = sorted(
                resolver.resolve_nameservers(name, resolve_addresses=resolve_addresses)
            )

    guarded_run(f, module, generate_additional_results=lambda: {"results": results})
    module.exit_json(results=results)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2024, Michael Ilg
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: consul_agent_service
short_description: Add, modify and delete services within a Consul cluster
version_added: 9.1.0
description:
  - Allows the addition, modification and deletion of services in a Consul cluster using the agent.
  - There are currently no plans to create services and checks in one. This is because the Consul API does not provide checks
    for a service and the checks themselves do not match the module parameters. Therefore, only a service without checks can
    be created in this module.
author:
  - Michael Ilg (@Ilgmi)
extends_documentation_fragment:
  - community.general._consul
  - community.general._consul.actiongroup_consul
  - community.general._consul.token
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: partial
    details:
      - In check mode the diff misses operational attributes.
options:
  state:
    description:
      - Whether the service should be present or absent.
    choices: ['present', 'absent']
    default: present
    type: str
  name:
    description:
      - Unique name for the service on a node, must be unique per node, required if registering a service.
    type: str
  id:
    description:
      - Specifies a unique ID for this service. This must be unique per agent. This defaults to the O(name) parameter if not
        provided. If O(state=absent), defaults to the service name if supplied.
    type: str
  tags:
    description:
      - Tags that are attached to the service registration.
    type: list
    elements: str
  address:
    description:
      - The address to advertise that the service listens on. This value is passed as the C(address) parameter to Consul's
        C(/v1/agent/service/register) API method, so refer to the Consul API documentation for further details.
    type: str
  meta:
    description:
      - Optional meta data used for filtering. For keys, the characters C(A-Z), C(a-z), C(0-9), C(_), C(-) are allowed. Not
        allowed characters are replaced with underscores.
    type: dict
  service_port:
    description:
      - The port on which the service is listening. Can optionally be supplied for registration of a service, that is if O(name)
        or O(id) is set.
    type: int
  enable_tag_override:
    description:
      - Specifies to disable the anti-entropy feature for this service's tags. If C(EnableTagOverride) is set to true then
        external agents can update this service in the catalog and modify the tags.
    type: bool
    default: false
  weights:
    description:
      - Specifies weights for the service.
    type: dict
    suboptions:
      passing:
        description:
          - Weights for passing.
        type: int
        default: 1
      warning:
        description:
          - Weights for warning.
        type: int
        default: 1
    default: {"passing": 1, "warning": 1}
"""

EXAMPLES = r"""
- name: Register nginx service with the local Consul agent
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    name: nginx
    service_port: 80

- name: Register nginx with a tcp check
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    name: nginx
    service_port: 80

- name: Register nginx with an http check
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    name: nginx
    service_port: 80

- name: Register external service nginx available at 10.1.5.23
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    name: nginx
    service_port: 80
    address: 10.1.5.23

- name: Register nginx with some service tags
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    name: nginx
    service_port: 80
    tags:
      - prod
      - webservers

- name: Register nginx with some service meta
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    name: nginx
    service_port: 80
    meta:
      nginx_version: 1.25.3

- name: Remove nginx service
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    service_id: nginx
    state: absent

- name: Register celery worker service
  community.general.consul_agent_service:
    host: consul1.example.com
    token: some_management_acl
    name: celery-worker
    tags:
      - prod
      - worker
"""

RETURN = r"""
service:
  description: The service as returned by the Consul HTTP API.
  returned: always
  type: dict
  sample:
    ID: nginx
    Service: nginx
    Address: localhost
    Port: 80
    Tags:
      - http
    Meta:
      - nginx_version: 1.23.3
    Datacenter: dc1
    Weights:
      Passing: 1
      Warning: 1
    ContentHash: 61a245cd985261ac
    EnableTagOverride: false
operation:
  description: The operation performed.
  returned: changed
  type: str
  sample: update
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._consul import (
    AUTH_ARGUMENTS_SPEC,
    OPERATION_CREATE,
    OPERATION_DELETE,
    OPERATION_UPDATE,
    _ConsulModule,
)

_CHECK_MUTUALLY_EXCLUSIVE = [("args", "ttl", "tcp", "http")]
_CHECK_REQUIRED_BY = {
    "args": "interval",
    "http": "interval",
    "tcp": "interval",
}

_ARGUMENT_SPEC = {
    "state": dict(default="present", choices=["present", "absent"]),
    "name": dict(type="str"),
    "id": dict(type="str"),
    "tags": dict(type="list", elements="str"),
    "address": dict(type="str"),
    "meta": dict(type="dict"),
    "service_port": dict(type="int"),
    "enable_tag_override": dict(type="bool", default=False),
    "weights": dict(
        type="dict",
        options=dict(passing=dict(type="int", default=1, no_log=False), warning=dict(type="int", default=1)),
        default={"passing": 1, "warning": 1},
    ),
}

_REQUIRED_IF = [
    ("state", "present", ["name"]),
    ("state", "absent", ("id", "name"), True),
]

_ARGUMENT_SPEC.update(AUTH_ARGUMENTS_SPEC)


class ConsulAgentServiceModule(_ConsulModule):
    api_endpoint = "agent/service"
    result_key = "service"
    unique_identifiers = ["id", "name"]
    operational_attributes = {"Service", "ContentHash", "Datacenter"}

    def endpoint_url(self, operation, identifier=None):
        if operation in [OPERATION_CREATE, OPERATION_UPDATE]:
            return f"{self.api_endpoint}/register"
        if operation == OPERATION_DELETE:
            return f"{self.api_endpoint}/deregister/{identifier}"

        return super().endpoint_url(operation, identifier)

    def prepare_object(self, existing, obj):
        existing = super().prepare_object(existing, obj)
        if "ServicePort" in existing:
            existing["Port"] = existing.pop("ServicePort")

        if "ID" not in existing:
            existing["ID"] = existing["Name"]

        return existing

    def needs_update(self, api_obj, module_obj):
        obj = {}
        if "Service" in api_obj:
            obj["Service"] = api_obj["Service"]
        api_obj = self.prepare_object(api_obj, obj)

        if "Name" in module_obj:
            module_obj["Service"] = module_obj.pop("Name")
        if "ServicePort" in module_obj:
            module_obj["Port"] = module_obj.pop("ServicePort")

        return super().needs_update(api_obj, module_obj)

    def delete_object(self, obj):
        if not self._module.check_mode:
            url = self.endpoint_url(OPERATION_DELETE, self.id_from_obj(obj, camel_case=True))
            self.put(url)
        return {}


def main():
    module = AnsibleModule(
        _ARGUMENT_SPEC,
        required_if=_REQUIRED_IF,
        supports_check_mode=True,
    )

    consul_module = ConsulAgentServiceModule(module)
    consul_module.execute()


if __name__ == "__main__":
    main()



# Copyright (c) 2016, Yanis Guenane <yanis+ansible@guenane.org>
# Copyright (c) 2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import binascii
import typing as t

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.crypto.plugins.module_utils._argspec import (
    ArgumentSpec,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLBadPassphraseError,
    OpenSSLObjectError,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.cryptography_crl import (
    REVOCATION_REASON_MAP,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.cryptography_support import (
    cryptography_get_basic_constraints,
    cryptography_get_name,
    cryptography_key_needs_digest_for_signing,
    cryptography_name_to_oid,
    cryptography_parse_key_usage_params,
    cryptography_parse_relative_distinguished_name,
    is_potential_certificate_issuer_public_key,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.module_backends.csr_info import (
    get_csr_info,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.support import (
    load_certificate_issuer_privatekey,
    load_certificate_request,
    parse_name_field,
    parse_ordered_name_field,
    select_message_digest,
)
from ansible_collections.community.crypto.plugins.module_utils._cryptography_dep import (
    COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION,
    assert_required_cryptography_version,
)

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover
    from cryptography.hazmat.primitives.asymmetric.types import (  # pragma: no cover
        CertificateIssuerPrivateKeyTypes,
    )

    _ET = t.TypeVar("_ET", bound="cryptography.x509.ExtensionType")  # pragma: no cover


MINIMAL_CRYPTOGRAPHY_VERSION = COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION

try:
    import cryptography
    import cryptography.exceptions
    import cryptography.hazmat.backends
    import cryptography.hazmat.primitives.hashes
    import cryptography.hazmat.primitives.serialization
    import cryptography.x509
    import cryptography.x509.oid
except ImportError:
    pass


class CertificateSigningRequestError(OpenSSLObjectError):
    pass


# From the object called `module`, only the following properties are used:
#
#  - module.params[]
#  - module.warn(msg: str)
#  - module.fail_json(msg: str, **kwargs)


def parse_crl_distribution_points(
    *, module: AnsibleModule, crl_distribution_points: list[dict[str, t.Any]]
) -> list[cryptography.x509.DistributionPoint]:
    result = []
    for index, parse_crl_distribution_point in enumerate(crl_distribution_points):
        try:
            full_name = None
            relative_name = None
            crl_issuer = None
            reasons = None
            if parse_crl_distribution_point["full_name"] is not None:
                if not parse_crl_distribution_point["full_name"]:
                    raise OpenSSLObjectError("full_name must not be empty")
                full_name = [
                    cryptography_get_name(name, what="full name")
                    for name in parse_crl_distribution_point["full_name"]
                ]
            if parse_crl_distribution_point["relative_name"] is not None:
                if not parse_crl_distribution_point["relative_name"]:
                    raise OpenSSLObjectError("relative_name must not be empty")
                relative_name = cryptography_parse_relative_distinguished_name(
                    parse_crl_distribution_point["relative_name"]
                )
            if parse_crl_distribution_point["crl_issuer"] is not None:
                if not parse_crl_distribution_point["crl_issuer"]:
                    raise OpenSSLObjectError("crl_issuer must not be empty")
                crl_issuer = [
                    cryptography_get_name(name, what="CRL issuer")
                    for name in parse_crl_distribution_point["crl_issuer"]
                ]
            if parse_crl_distribution_point["reasons"] is not None:
                reasons_list = []
                for reason in parse_crl_distribution_point["reasons"]:
                    reasons_list.append(REVOCATION_REASON_MAP[reason])
                reasons = frozenset(reasons_list)
            result.append(
                cryptography.x509.DistributionPoint(
                    full_name=full_name,
                    relative_name=relative_name,
                    crl_issuer=crl_issuer,
                    reasons=reasons,
                )
            )
        except (OpenSSLObjectError, ValueError) as e:
            raise OpenSSLObjectError(
                f"Error while parsing CRL distribution point #{index}: {e}"
            ) from e
    return result


class CertificateSigningRequestBackend:
    def __init__(self, *, module: AnsibleModule) -> None:
        self.module = module
        self.digest: str = module.params["digest"]
        self.privatekey_path: str | None = module.params["privatekey_path"]
        privatekey_content: str | None = module.params["privatekey_content"]
        if privatekey_content is not None:
            self.privatekey_content: bytes | None = privatekey_content.encode("utf-8")
        else:
            self.privatekey_content = None
        self.privatekey_passphrase: str | None = module.params["privatekey_passphrase"]
        self.version: t.Literal[1] = module.params["version"]
        self.subject_alt_name: list[str] | None = module.params["subject_alt_name"]
        self.subject_alt_name_critical: bool = module.params[
            "subject_alt_name_critical"
        ]
        self.key_usage: list[str] | None = module.params["key_usage"]
        self.key_usage_critical: bool = module.params["key_usage_critical"]
        self.extended_key_usage: list[str] | None = module.params["extended_key_usage"]
        self.extended_key_usage_critical: bool = module.params[
            "extended_key_usage_critical"
        ]
        self.basic_constraints: list[str] | None = module.params["basic_constraints"]
        self.basic_constraints_critical: bool = module.params[
            "basic_constraints_critical"
        ]
        self.ocsp_must_staple: bool = module.params["ocsp_must_staple"]
        self.ocsp_must_staple_critical: bool = module.params[
            "ocsp_must_staple_critical"
        ]
        self.name_constraints_permitted: list[str] = (
            module.params["name_constraints_permitted"] or []
        )
        self.name_constraints_excluded: list[str] = (
            module.params["name_constraints_excluded"] or []
        )
        self.name_constraints_critical: bool = module.params[
            "name_constraints_critical"
        ]
        self.create_subject_key_identifier: bool = module.params[
            "create_subject_key_identifier"
        ]
        subject_key_identifier: str | None = module.params["subject_key_identifier"]
        authority_key_identifier: str | None = module.params["authority_key_identifier"]
        self.authority_cert_issuer: list[str] | None = module.params[
            "authority_cert_issuer"
        ]
        self.authority_cert_serial_number: int | None = module.params[
            "authority_cert_serial_number"
        ]
        self.crl_distribution_points: (
            list[cryptography.x509.DistributionPoint] | None
        ) = None
        self.csr: cryptography.x509.CertificateSigningRequest | None = None
        self.privatekey: CertificateIssuerPrivateKeyTypes | None = None

        if self.create_subject_key_identifier and subject_key_identifier is not None:
            module.fail_json(
                msg="subject_key_identifier cannot be specified if create_subject_key_identifier is true"
            )

        self.ordered_subject = False
        subject = [
            ("C", module.params["country_name"]),
            ("ST", module.params["state_or_province_name"]),
            ("L", module.params["locality_name"]),
            ("O", module.params["organization_name"]),
            ("OU", module.params["organizational_unit_name"]),
            ("CN", module.params["common_name"]),
            ("emailAddress", module.params["email_address"]),
        ]
        self.subject: list[tuple[str, str]] = [
            (entry[0], entry[1]) for entry in subject if entry[1]
        ]

        try:
            if module.params["subject"]:
                self.subject = self.subject + parse_name_field(
                    module.params["subject"], name_field_name="subject"
                )
            if module.params["subject_ordered"]:
                if self.subject:
                    raise CertificateSigningRequestError(
                        "subject_ordered cannot be combined with any other subject field"
                    )
                self.subject = parse_ordered_name_field(
                    module.params["subject_ordered"], name_field_name="subject_ordered"
                )
                self.ordered_subject = True
        except ValueError as exc:
            raise CertificateSigningRequestError(str(exc)) from exc

        self.using_common_name_for_san = False
        if not self.subject_alt_name and module.params["use_common_name_for_san"]:
            for sub in self.subject:
                if sub[0] in ("commonName", "CN"):
                    self.subject_alt_name = [f"DNS:{sub[1]}"]
                    self.using_common_name_for_san = True
                    break

        self.subject_key_identifier: bytes | None = None
        if subject_key_identifier is not None:
            try:
                self.subject_key_identifier = binascii.unhexlify(
                    subject_key_identifier.replace(":", "")
                )
            except Exception as e:
                raise CertificateSigningRequestError(
                    f"Cannot parse subject_key_identifier: {e}"
                ) from e

        self.authority_key_identifier: bytes | None = None
        if authority_key_identifier is not None:
            try:
                self.authority_key_identifier = binascii.unhexlify(
                    authority_key_identifier.replace(":", "")
                )
            except Exception as e:
                raise CertificateSigningRequestError(
                    f"Cannot parse authority_key_identifier: {e}"
                ) from e

        self.existing_csr: cryptography.x509.CertificateSigningRequest | None = None
        self.existing_csr_bytes: bytes | None = None

        self.diff_before = self._get_info(data=None)
        self.diff_after = self._get_info(data=None)

        crl_distribution_points: list[dict[str, t.Any]] | None = module.params[
            "crl_distribution_points"
        ]
        if crl_distribution_points:
            self.crl_distribution_points = parse_crl_distribution_points(
                module=module, crl_distribution_points=crl_distribution_points
            )

    def _get_info(self, *, data: bytes | None) -> dict[str, t.Any]:
        if data is None:
            return {}
        try:
            result = get_csr_info(
                module=self.module,
                content=data,
                validate_signature=False,
                prefer_one_fingerprint=True,
            )
            result["can_parse_csr"] = True
            return result
        except Exception:
            return {"can_parse_csr": False}

    def generate_csr(self) -> None:
        """(Re-)Generate CSR."""
        self._ensure_private_key_loaded()
        assert self.privatekey is not None

        csr = cryptography.x509.CertificateSigningRequestBuilder()
        try:
            csr = csr.subject_name(
                cryptography.x509.Name(
                    [
                        cryptography.x509.NameAttribute(
                            cryptography_name_to_oid(entry[0]), to_text(entry[1])
                        )
                        for entry in self.subject
                    ]
                )
            )
        except ValueError as e:
            raise CertificateSigningRequestError(e) from e

        if self.subject_alt_name:
            csr = csr.add_extension(
                cryptography.x509.SubjectAlternativeName(
                    [cryptography_get_name(name) for name in self.subject_alt_name]
                ),
                critical=self.subject_alt_name_critical,
            )

        if self.key_usage:
            params = cryptography_parse_key_usage_params(self.key_usage)
            csr = csr.add_extension(
                cryptography.x509.KeyUsage(**params), critical=self.key_usage_critical
            )

        if self.extended_key_usage:
            usages = [
                cryptography_name_to_oid(usage) for usage in self.extended_key_usage
            ]
            csr = csr.add_extension(
                cryptography.x509.ExtendedKeyUsage(usages),
                critical=self.extended_key_usage_critical,
            )

        if self.basic_constraints:
            params = {}
            ca, path_length = cryptography_get_basic_constraints(self.basic_constraints)
            csr = csr.add_extension(
                cryptography.x509.BasicConstraints(ca, path_length),
                critical=self.basic_constraints_critical,
            )

        if self.ocsp_must_staple:
            csr = csr.add_extension(
                cryptography.x509.TLSFeature(
                    [cryptography.x509.TLSFeatureType.status_request]
                ),
                critical=self.ocsp_must_staple_critical,
            )

        if self.name_constraints_permitted or self.name_constraints_excluded:
            try:
                csr = csr.add_extension(
                    cryptography.x509.NameConstraints(
                        [
                            cryptography_get_name(
                                name, what="name constraints permitted"
                            )
                            for name in self.name_constraints_permitted
                        ]
                        or None,
                        [
                            cryptography_get_name(
                                name, what="name constraints excluded"
                            )
                            for name in self.name_constraints_excluded
                        ]
                        or None,
                    ),
                    critical=self.name_constraints_critical,
                )
            except TypeError as e:
                raise OpenSSLObjectError(
                    f"Error while parsing name constraint: {e}"
                ) from e

        if self.create_subject_key_identifier:
            if not is_potential_certificate_issuer_public_key(
                self.privatekey.public_key()
            ):
                raise OpenSSLObjectError(
                    "Private key can not be used to create subject key identifier"
                )
            csr = csr.add_extension(
                cryptography.x509.SubjectKeyIdentifier.from_public_key(
                    self.privatekey.public_key()
                ),
                critical=False,
            )
        elif self.subject_key_identifier is not None:
            csr = csr.add_extension(
                cryptography.x509.SubjectKeyIdentifier(self.subject_key_identifier),
                critical=False,
            )

        if (
            self.authority_key_identifier is not None
            or self.authority_cert_issuer is not None
            or self.authority_cert_serial_number is not None
        ):
            issuers = None
            if self.authority_cert_issuer is not None:
                issuers = [
                    cryptography_get_name(n, what="authority cert issuer")
                    for n in self.authority_cert_issuer
                ]
            csr = csr.add_extension(
                cryptography.x509.AuthorityKeyIdentifier(
                    self.authority_key_identifier,
                    issuers,
                    self.authority_cert_serial_number,
                ),
                critical=False,
            )

        if self.crl_distribution_points:
            csr = csr.add_extension(
                cryptography.x509.CRLDistributionPoints(self.crl_distribution_points),
                critical=False,
            )

        # csr.sign() does not accept some digests we theoretically could have in digest.
        # For that reason we use type t.Any here. csr.sign() will complain if
        # the digest is not acceptable.
        digest: t.Any | None = None
        if cryptography_key_needs_digest_for_signing(self.privatekey):
            digest = select_message_digest(self.digest)
            if digest is None:
                raise CertificateSigningRequestError(
                    f'Unsupported digest "{self.digest}"'
                )
        try:
            self.csr = csr.sign(self.privatekey, digest)
        except UnicodeError as e:
            # This catches IDNAErrors, which happens when a bad name is passed as a SAN
            # (https://github.com/ansible-collections/community.crypto/issues/105).
            # For older cryptography versions, this is handled by idna, which raises
            # an idna.core.IDNAError. Later versions of cryptography deprecated and stopped
            # requiring idna, whence we cannot easily handle this error. Fortunately, in
            # most versions of idna, IDNAError extends UnicodeError. There is only version
            # 2.3 where it extends Exception instead (see
            # https://github.com/kjd/idna/commit/ebefacd3134d0f5da4745878620a6a1cba86d130
            # and then
            # https://github.com/kjd/idna/commit/ea03c7b5db7d2a99af082e0239da2b68aeea702a).
            msg = f"Error while creating CSR: {e}\n"
            if self.using_common_name_for_san:
                self.module.fail_json(
                    msg=msg
                    + "This is probably caused because the Common Name is used as a SAN. Specifying use_common_name_for_san=false might fix this."
                )
            self.module.fail_json(
                msg=msg
                + "This is probably caused by an invalid Subject Alternative DNS Name."
            )

    def get_csr_data(self) -> bytes:
        """Return bytes for self.csr."""
        if self.csr is None:
            raise AssertionError(
                "Violated contract: csr is not populated"
            )  # pragma: no cover
        return self.csr.public_bytes(
            cryptography.hazmat.primitives.serialization.Encoding.PEM
        )

    def set_existing(self, *, csr_bytes: bytes | None) -> None:
        """Set existing CSR bytes. None indicates that the CSR does not exist."""
        self.existing_csr_bytes = csr_bytes
        self.diff_after = self.diff_before = self._get_info(
            data=self.existing_csr_bytes
        )

    def has_existing(self) -> bool:
        """Query whether an existing CSR is/has been there."""
        return self.existing_csr_bytes is not None

    def _ensure_private_key_loaded(self) -> None:
        """Load the provided private key into self.privatekey."""
        if self.privatekey is not None:
            return
        try:
            self.privatekey = load_certificate_issuer_privatekey(
                path=self.privatekey_path,
                content=self.privatekey_content,
                passphrase=self.privatekey_passphrase,
            )
        except OpenSSLBadPassphraseError as exc:
            raise CertificateSigningRequestError(exc) from exc

    def _check_csr(self) -> bool:
        """Check whether provided parameters, assuming self.existing_csr and self.privatekey have been populated."""
        if self.existing_csr is None:
            raise AssertionError(
                "Violated contract: existing_csr is not populated"
            )  # pragma: no cover
        if self.privatekey is None:
            raise AssertionError(
                "Violated contract: privatekey is not populated"
            )  # pragma: no cover

        def _check_subject(csr: cryptography.x509.CertificateSigningRequest) -> bool:
            subject = [
                (cryptography_name_to_oid(entry[0]), to_text(entry[1]))
                for entry in self.subject
            ]
            current_subject = [(sub.oid, sub.value) for sub in csr.subject]
            if self.ordered_subject:
                return subject == current_subject
            return set(subject) == set(current_subject)

        def _find_extension(
            extensions: cryptography.x509.Extensions, exttype: type[_ET]
        ) -> cryptography.x509.Extension[_ET] | None:
            return next(
                (ext for ext in extensions if isinstance(ext.value, exttype)), None
            )

        def _check_subject_alt_name(extensions: cryptography.x509.Extensions) -> bool:
            current_altnames_ext = _find_extension(
                extensions, cryptography.x509.SubjectAlternativeName
            )
            current_altnames = (
                [to_text(altname) for altname in current_altnames_ext.value]
                if current_altnames_ext
                else []
            )
            altnames = (
                [
                    to_text(cryptography_get_name(altname))
                    for altname in self.subject_alt_name
                ]
                if self.subject_alt_name
                else []
            )
            if set(altnames) != set(current_altnames):
                return False
            return not (
                altnames
                and current_altnames_ext
                and current_altnames_ext.critical != self.subject_alt_name_critical
            )

        def _check_key_usage(extensions: cryptography.x509.Extensions) -> bool:
            current_keyusage_ext = _find_extension(
                extensions, cryptography.x509.KeyUsage
            )
            if not self.key_usage:
                return current_keyusage_ext is None
            if current_keyusage_ext is None:
                return False
            params = cryptography_parse_key_usage_params(self.key_usage)
            for param, value in params.items():
                try:
                    # param in ('encipher_only', 'decipher_only') can result in ValueError()
                    # being raised if key_agreement == False.
                    current_value = getattr(current_keyusage_ext.value, param)
                except ValueError:
                    # In that case, assume that the value is False.
                    current_value = False
                if current_value != value:
                    return False
            return current_keyusage_ext.critical == self.key_usage_critical

        def _check_extended_key_usage(extensions: cryptography.x509.Extensions) -> bool:
            current_usages_ext = _find_extension(
                extensions, cryptography.x509.ExtendedKeyUsage
            )
            current_usages = (
                [str(usage) for usage in current_usages_ext.value]
                if current_usages_ext
                else []
            )
            usages = (
                [
                    str(cryptography_name_to_oid(usage))
                    for usage in self.extended_key_usage
                ]
                if self.extended_key_usage
                else []
            )
            if set(current_usages) != set(usages):
                return False
            return not (
                usages
                and current_usages_ext
                and current_usages_ext.critical != self.extended_key_usage_critical
            )

        def _check_basic_constraints(extensions: cryptography.x509.Extensions) -> bool:
            bc_ext = _find_extension(extensions, cryptography.x509.BasicConstraints)
            current_ca = bc_ext.value.ca if bc_ext else False
            current_path_length = bc_ext.value.path_length if bc_ext else None
            ca, path_length = cryptography_get_basic_constraints(self.basic_constraints)
            # Check CA flag
            if ca != current_ca:
                return False
            # Check path length
            if path_length != current_path_length:
                return False
            # Check criticality
            if self.basic_constraints:
                return (
                    bc_ext is not None
                    and bc_ext.critical == self.basic_constraints_critical
                )
            return bc_ext is None

        def _check_ocsp_must_staple(extensions: cryptography.x509.Extensions) -> bool:
            tlsfeature_ext = _find_extension(extensions, cryptography.x509.TLSFeature)
            if self.ocsp_must_staple:
                if (
                    not tlsfeature_ext
                    or tlsfeature_ext.critical != self.ocsp_must_staple_critical
                ):
                    return False
                return (
                    cryptography.x509.TLSFeatureType.status_request
                    in tlsfeature_ext.value
                )
            return tlsfeature_ext is None

        def _check_name_constraints(extensions: cryptography.x509.Extensions) -> bool:
            current_nc_ext = _find_extension(
                extensions, cryptography.x509.NameConstraints
            )
            current_nc_perm = (
                [
                    to_text(altname)
                    for altname in current_nc_ext.value.permitted_subtrees or []
                ]
                if current_nc_ext
                else []
            )
            current_nc_excl = (
                [
                    to_text(altname)
                    for altname in current_nc_ext.value.excluded_subtrees or []
                ]
                if current_nc_ext
                else []
            )
            nc_perm = [
                to_text(
                    cryptography_get_name(altname, what="name constraints permitted")
                )
                for altname in self.name_constraints_permitted
            ]
            nc_excl = [
                to_text(
                    cryptography_get_name(altname, what="name constraints excluded")
                )
                for altname in self.name_constraints_excluded
            ]
            if set(nc_perm) != set(current_nc_perm) or set(nc_excl) != set(
                current_nc_excl
            ):
                return False
            return not (
                (nc_perm or nc_excl)
                and current_nc_ext
                and current_nc_ext.critical != self.name_constraints_critical
            )

        def _check_subject_key_identifier(
            extensions: cryptography.x509.Extensions,
        ) -> bool:
            ext = _find_extension(extensions, cryptography.x509.SubjectKeyIdentifier)
            if (
                self.create_subject_key_identifier
                or self.subject_key_identifier is not None
            ):
                if not ext or ext.critical:
                    return False
                if self.create_subject_key_identifier:
                    assert self.privatekey is not None
                    digest = cryptography.x509.SubjectKeyIdentifier.from_public_key(
                        self.privatekey.public_key()
                    ).digest
                    return ext.value.digest == digest
                return ext.value.digest == self.subject_key_identifier
            return ext is None

        def _check_authority_key_identifier(
            extensions: cryptography.x509.Extensions,
        ) -> bool:
            ext = _find_extension(extensions, cryptography.x509.AuthorityKeyIdentifier)
            if (
                self.authority_key_identifier is not None
                or self.authority_cert_issuer is not None
                or self.authority_cert_serial_number is not None
            ):
                if not ext or ext.critical:
                    return False
                aci = None
                csr_aci = None
                if self.authority_cert_issuer is not None:
                    aci = [
                        to_text(cryptography_get_name(n, what="authority cert issuer"))
                        for n in self.authority_cert_issuer
                    ]
                if ext.value.authority_cert_issuer is not None:
                    csr_aci = [to_text(n) for n in ext.value.authority_cert_issuer]
                return (
                    ext.value.key_identifier == self.authority_key_identifier
                    and csr_aci == aci
                    and ext.value.authority_cert_serial_number
                    == self.authority_cert_serial_number
                )
            return ext is None

        def _check_crl_distribution_points(
            extensions: cryptography.x509.Extensions,
        ) -> bool:
            ext = _find_extension(extensions, cryptography.x509.CRLDistributionPoints)
            if self.crl_distribution_points is None:
                return ext is None
            if not ext:
                return False
            return list(ext.value) == self.crl_distribution_points

        def _check_extensions(csr: cryptography.x509.CertificateSigningRequest) -> bool:
            extensions = csr.extensions
            return (
                _check_subject_alt_name(extensions)
                and _check_key_usage(extensions)
                and _check_extended_key_usage(extensions)
                and _check_basic_constraints(extensions)
                and _check_ocsp_must_staple(extensions)
                and _check_subject_key_identifier(extensions)
                and _check_authority_key_identifier(extensions)
                and _check_name_constraints(extensions)
                and _check_crl_distribution_points(extensions)
            )

        def _check_signature(csr: cryptography.x509.CertificateSigningRequest) -> bool:
            if not csr.is_signature_valid:
                return False
            # To check whether public key of CSR belongs to private key,
            # encode both public keys and compare PEMs.
            key_a = csr.public_key().public_bytes(
                cryptography.hazmat.primitives.serialization.Encoding.PEM,
                cryptography.hazmat.primitives.serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            assert self.privatekey is not None
            key_b = self.privatekey.public_key().public_bytes(
                cryptography.hazmat.primitives.serialization.Encoding.PEM,
                cryptography.hazmat.primitives.serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            return key_a == key_b

        return (
            _check_subject(self.existing_csr)
            and _check_extensions(self.existing_csr)
            and _check_signature(self.existing_csr)
        )

    def needs_regeneration(self) -> bool:
        """Check whether a regeneration is necessary."""
        if self.existing_csr_bytes is None:
            return True
        try:
            self.existing_csr = load_certificate_request(
                content=self.existing_csr_bytes,
            )
        except Exception:
            return True
        self._ensure_private_key_loaded()
        return not self._check_csr()

    def dump(self, *, include_csr: bool) -> dict[str, t.Any]:
        """Serialize the object into a dictionary."""
        result: dict[str, t.Any] = {
            "privatekey": self.privatekey_path,
            "subject": self.subject,
            "subjectAltName": self.subject_alt_name,
            "keyUsage": self.key_usage,
            "extendedKeyUsage": self.extended_key_usage,
            "basicConstraints": self.basic_constraints,
            "ocspMustStaple": self.ocsp_must_staple,
            "name_constraints_permitted": self.name_constraints_permitted,
            "name_constraints_excluded": self.name_constraints_excluded,
        }
        # Get hold of CSR bytes
        csr_bytes = self.existing_csr_bytes
        if self.csr is not None:
            csr_bytes = self.get_csr_data()
        self.diff_after = self._get_info(data=csr_bytes)
        if include_csr:
            # Store result
            result["csr"] = csr_bytes.decode("utf-8") if csr_bytes else None

        result["diff"] = {
            "before": self.diff_before,
            "after": self.diff_after,
        }
        return result


def select_backend(
    module: AnsibleModule,
) -> CertificateSigningRequestBackend:
    assert_required_cryptography_version(
        module, minimum_cryptography_version=MINIMAL_CRYPTOGRAPHY_VERSION
    )
    return CertificateSigningRequestBackend(module=module)


def get_csr_argument_spec() -> ArgumentSpec:
    return ArgumentSpec(
        argument_spec={
            "digest": {"type": "str", "default": "sha256"},
            "privatekey_path": {"type": "path"},
            "privatekey_content": {"type": "str", "no_log": True},
            "privatekey_passphrase": {"type": "str", "no_log": True},
            "version": {"type": "int", "default": 1, "choices": [1]},
            "subject": {"type": "dict"},
            "subject_ordered": {"type": "list", "elements": "dict"},
            "country_name": {"type": "str", "aliases": ["C", "countryName"]},
            "state_or_province_name": {
                "type": "str",
                "aliases": ["ST", "stateOrProvinceName"],
            },
            "locality_name": {"type": "str", "aliases": ["L", "localityName"]},
            "organization_name": {"type": "str", "aliases": ["O", "organizationName"]},
            "organizational_unit_name": {
                "type": "str",
                "aliases": ["OU", "organizationalUnitName"],
            },
            "common_name": {"type": "str", "aliases": ["CN", "commonName"]},
            "email_address": {"type": "str", "aliases": ["E", "emailAddress"]},
            "subject_alt_name": {
                "type": "list",
                "elements": "str",
                "aliases": ["subjectAltName"],
            },
            "subject_alt_name_critical": {
                "type": "bool",
                "default": False,
                "aliases": ["subjectAltName_critical"],
            },
            "use_common_name_for_san": {
                "type": "bool",
                "default": True,
                "aliases": ["useCommonNameForSAN"],
            },
            "key_usage": {"type": "list", "elements": "str", "aliases": ["keyUsage"]},
            "key_usage_critical": {
                "type": "bool",
                "default": False,
                "aliases": ["keyUsage_critical"],
            },
            "extended_key_usage": {
                "type": "list",
                "elements": "str",
                "aliases": ["extKeyUsage", "extendedKeyUsage"],
            },
            "extended_key_usage_critical": {
                "type": "bool",
                "default": False,
                "aliases": ["extKeyUsage_critical", "extendedKeyUsage_critical"],
            },
            "basic_constraints": {
                "type": "list",
                "elements": "str",
                "aliases": ["basicConstraints"],
            },
            "basic_constraints_critical": {
                "type": "bool",
                "default": False,
                "aliases": ["basicConstraints_critical"],
            },
            "ocsp_must_staple": {
                "type": "bool",
                "default": False,
                "aliases": ["ocspMustStaple"],
            },
            "ocsp_must_staple_critical": {
                "type": "bool",
                "default": False,
                "aliases": ["ocspMustStaple_critical"],
            },
            "name_constraints_permitted": {"type": "list", "elements": "str"},
            "name_constraints_excluded": {"type": "list", "elements": "str"},
            "name_constraints_critical": {"type": "bool", "default": False},
            "create_subject_key_identifier": {"type": "bool", "default": False},
            "subject_key_identifier": {"type": "str"},
            "authority_key_identifier": {"type": "str"},
            "authority_cert_issuer": {"type": "list", "elements": "str"},
            "authority_cert_serial_number": {"type": "int"},
            "crl_distribution_points": {
                "type": "list",
                "elements": "dict",
                "options": {
                    "full_name": {"type": "list", "elements": "str"},
                    "relative_name": {"type": "list", "elements": "str"},
                    "crl_issuer": {"type": "list", "elements": "str"},
                    "reasons": {
                        "type": "list",
                        "elements": "str",
                        "choices": [
                            "key_compromise",
                            "ca_compromise",
                            "affiliation_changed",
                            "superseded",
                            "cessation_of_operation",
                            "certificate_hold",
                            "privilege_withdrawn",
                            "aa_compromise",
                        ],
                    },
                },
                "mutually_exclusive": [("full_name", "relative_name")],
                "required_one_of": [("full_name", "relative_name", "crl_issuer")],
            },
            "select_crypto_backend": {
                "type": "str",
                "default": "auto",
                "choices": ["auto", "cryptography"],
            },
        },
        required_together=[
            ["authority_cert_issuer", "authority_cert_serial_number"],
        ],
        mutually_exclusive=[
            ["privatekey_path", "privatekey_content"],
            ["subject", "subject_ordered"],
        ],
        required_one_of=[
            ["privatekey_path", "privatekey_content"],
        ],
    )


__all__ = (
    "CertificateSigningRequestBackend",
    "CertificateSigningRequestError",
    "get_csr_argument_spec",
    "select_backend",
)



# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2026 Christoph Fiehe <christoph.fiehe@gmail.com>

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Use together with ansible.builtin.url and icinga2_argument_spec from
    # ansible_collections.community.general.plugins.module_utils._icinga2
    DOCUMENTATION = r"""
options:
  url:
    description:
      - URL of the Icinga 2 REST API.
    type: str
    required: true
  ca_path:
    description:
      - CA certificates bundle to use to verify the Icinga 2 server certificate.
    type: path
  timeout:
    description:
      - How long to wait for the server to send data before giving up.
    type: int
    default: 10
"""



# Copyright (c) Ansible project
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import json
import typing as t
from http import HTTPStatus

from ansible.module_utils.basic import env_fallback
from ansible.module_utils.urls import basic_auth_header, fetch_url

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


class BitbucketHelper:
    BITBUCKET_API_URL = "https://api.bitbucket.org"

    def __init__(self, module: AnsibleModule) -> None:
        self.module = module
        self.access_token: str | None = None

    @staticmethod
    def bitbucket_argument_spec() -> dict[str, t.Any]:
        return dict(
            client_id=dict(type="str", fallback=(env_fallback, ["BITBUCKET_CLIENT_ID"])),
            client_secret=dict(type="str", no_log=True, fallback=(env_fallback, ["BITBUCKET_CLIENT_SECRET"])),
            # TODO:
            # - Rename user to username once current usage of username is removed
            # - Alias user to username and deprecate it
            user=dict(type="str", aliases=["username"], fallback=(env_fallback, ["BITBUCKET_USERNAME"])),
            password=dict(type="str", no_log=True, fallback=(env_fallback, ["BITBUCKET_PASSWORD"])),
        )

    @staticmethod
    def bitbucket_required_one_of() -> list[list[str]]:
        return [["client_id", "client_secret", "user", "password"]]

    @staticmethod
    def bitbucket_required_together() -> list[list[str]]:
        return [["client_id", "client_secret"], ["user", "password"]]

    def fetch_access_token(self) -> None:
        if self.module.params["client_id"] and self.module.params["client_secret"]:
            headers = {
                "Authorization": basic_auth_header(
                    self.module.params["client_id"], self.module.params["client_secret"]
                ),
            }

            info, content = self.request(
                api_url="https://bitbucket.org/site/oauth2/access_token",
                method="POST",
                data="grant_type=client_credentials",
                headers=headers,
            )

            if info["status"] == HTTPStatus.OK:
                self.access_token = content["access_token"]
            else:
                self.module.fail_json(msg=f"Failed to retrieve access token: {info}")

    def request(
        self, api_url: str, method: str, data: bytes | str | dict | None = None, headers: dict[str, str] | None = None
    ):
        headers = headers or {}

        if self.access_token:
            headers.update(
                {
                    "Authorization": f"Bearer {self.access_token}",
                }
            )
        elif self.module.params["user"] and self.module.params["password"]:
            headers.update(
                {
                    "Authorization": basic_auth_header(self.module.params["user"], self.module.params["password"]),
                }
            )

        if isinstance(data, dict):
            data = self.module.jsonify(data)
            headers.update(
                {
                    "Content-type": "application/json",
                }
            )

        response, info = fetch_url(
            module=self.module,
            url=api_url,
            method=method,
            headers=headers,
            data=data,
            force=True,
        )

        content = {}

        if response is not None:
            body = response.read()
            if body:
                content = json.loads(body)

        return info, content



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_devices_availabilities_info
short_description: Information module for organizations _devices _availabilities
description:
  - Get all organizations _devices _availabilities. - > List the availability information
    for devices in an organization. The data returned by this endpoint is updated
    every 5 minutes.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  perPage:
    description:
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 1000. Default is 1000.
    type: int
  startingAfter:
    description:
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
  networkIds:
    description:
      - >
        NetworkIds query parameter. Optional parameter to filter device availabilities
        by network ID. This filter uses multiple exact matches.
    elements: str
    type: list
  productTypes:
    description:
      - >
        ProductTypes query parameter. Optional parameter to filter device availabilities
        by device product types. This filter uses multiple exact matches. Valid types
        are wireless, appliance, switch, camera, cellularGateway, sensor, wirelessController,
        and campusGateway.
    elements: str
    type: list
  serials:
    description:
      - >
        Serials query parameter. Optional parameter to filter device availabilities
        by device serial numbers. This filter uses multiple exact matches.
    elements: str
    type: list
  tags:
    description:
      - >
        Tags query parameter. An optional parameter to filter devices by tags. The
        filtering is case-sensitive. If tags are included, 'tagsFilterType' should
        also be included (see below). This filter uses multiple exact matches.
    elements: str
    type: list
  tagsFilterType:
    description:
      - >
        TagsFilterType query parameter. An optional parameter of value 'withAnyTags'
        or 'withAllTags' to indicate whether to return devices which contain ANY or
        ALL of the included tags. If no type is included, 'withAnyTags' will be selected.
    type: str
  statuses:
    description:
      - >
        Statuses query parameter. Optional parameter to filter device availabilities
        by device status. This filter uses multiple exact matches.
    elements: str
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations getOrganizationDevicesAvailabilities
    description: Complete reference of the getOrganizationDevicesAvailabilities API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-devices-availabilities
notes:
  - SDK Method used are
    organizations.Organizations.get_organization_devices_availabilities,
  - Paths used are
    get /organizations/{organizationId}/devices/availabilities,
"""

EXAMPLES = r"""
- name: Get all organizations _devices _availabilities
  cisco.meraki.organizations_devices_availabilities_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    perPage: 0
    startingAfter: string
    endingBefore: string
    networkIds: []
    productTypes: []
    serials: []
    tags: []
    tagsFilterType: string
    statuses: []
    organizationId: string
    total_pages: -1
    direction: next
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "mac": "string",
        "name": "string",
        "network": {
          "id": "string"
        },
        "productType": "string",
        "serial": "string",
        "status": "string",
        "tags": [
          "string"
        ]
      }
    ]
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: administered_licensing_subscription_subscriptions_claim
short_description: Resource module for administered _licensing _subscription _subscriptions
  _claim
description:
  - Manage operation create of the resource administered _licensing _subscription
    _subscriptions _claim.
  - Claim a subscription into an organization.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  claimKey:
    description: The subscription's claim key.
    type: str
  description:
    description: Extra details or notes about the subscription.
    type: str
  name:
    description: Friendly name to identify the subscription.
    type: str
  organizationId:
    description: The id of the organization claiming the subscription.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for licensing claimAdministeredLicensingSubscriptionSubscriptions
    description: Complete reference of the claimAdministeredLicensingSubscriptionSubscriptions
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!claim-administered-licensing-subscription-subscriptions
notes:
  - SDK Method used are
    licensing.Licensing.claim_administered_licensing_subscription_subscriptions,
  - Paths used are
    post /administered/licensing/subscription/subscriptions/claim,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.administered_licensing_subscription_subscriptions_claim:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    claimKey: S2345-6789A-BCDEF-GHJKM
    description: Subscription for all main offices
    name: Corporate subscription
    organizationId: '12345678910'
    validate: true
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "subscriptionId": "string",
      "name": "string",
      "description": "string",
      "status": "string",
      "startDate": "string",
      "endDate": "string",
      "lastUpdatedAt": "string",
      "webOrderId": "string",
      "type": "string",
      "smartAccount": {
        "status": "string",
        "account": {
          "id": "string",
          "name": "string",
          "domain": "string"
        }
      },
      "renewalRequested": true,
      "productTypes": [
        "string"
      ],
      "entitlements": [
        {
          "sku": "string",
          "seats": {
            "assigned": 0,
            "available": 0,
            "limit": 0
          }
        }
      ],
      "counts": {
        "seats": {
          "assigned": 0,
          "available": 0,
          "limit": 0
        },
        "networks": 0,
        "organizations": 0
      },
      "enterpriseAgreement": {
        "suites": [
          "string"
        ]
      }
    }
"""



#!/usr/bin/python

# Copyright (c) 2017, Eike Frost <ei@kefro.st>
# Copyright (c) 2021, Christophe Gilles <christophe.gilles54@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_authz_permission

version_added: 7.2.0

short_description: Allows administration of Keycloak client authorization permissions using Keycloak API

description:
  - This module allows the administration of Keycloak client authorization permissions using the Keycloak REST API. Authorization
    permissions are only available if a client has Authorization enabled.
  - There are some peculiarities in JSON paths and payloads for authorization permissions. In particular POST and PUT operations
    are targeted at permission endpoints, whereas GET requests go to policies endpoint. To make matters more interesting the
    JSON responses from GET requests return data in a different format than what is expected for POST and PUT. The end result
    is that it is not possible to detect changes to things like policies, scopes or resources - at least not without a large
    number of additional API calls. Therefore this module always updates authorization permissions instead of attempting to
    determine if changes are truly needed.
  - This module requires access to the REST API using OpenID Connect; the user connecting and the realm being used must have
    the requisite access rights. In a default Keycloak installation, admin-cli and an admin user would work, as would a separate
    realm definition with the scope tailored to your needs and a user having the expected roles.
  - The names of module options are snake_cased versions of the camelCase options used by Keycloak. The Authorization Services
    paths and payloads have not officially been documented by the Keycloak project.
    U(https://www.puppeteers.net/blog/keycloak-authorization-services-rest-api-paths-and-payload/).
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the authorization permission.
      - On V(present), the authorization permission is created (or updated if it exists already).
      - On V(absent), the authorization permission is removed if it exists.
    choices: ['present', 'absent']
    default: 'present'
    type: str
  name:
    description:
      - Name of the authorization permission to create.
    type: str
    required: true
  description:
    description:
      - The description of the authorization permission.
    type: str
  permission_type:
    description:
      - The type of authorization permission.
      - On V(scope) create a scope-based permission.
      - On V(resource) create a resource-based permission.
    type: str
    required: true
    choices:
      - resource
      - scope
  decision_strategy:
    description:
      - The decision strategy to use with this permission.
    type: str
    default: UNANIMOUS
    choices:
      - UNANIMOUS
      - AFFIRMATIVE
      - CONSENSUS
  resources:
    description:
      - Resource names to attach to this permission.
      - Scope-based permissions can only include one resource.
      - Resource-based permissions can include multiple resources.
    type: list
    elements: str
    default: []
  scopes:
    description:
      - Scope names to attach to this permission.
      - Resource-based permissions cannot have scopes attached to them.
    type: list
    elements: str
    default: []
  policies:
    description:
      - Policy names to attach to this permission.
    type: list
    elements: str
    default: []
  client_id:
    description:
      - The clientId of the keycloak client that should have the authorization scope.
      - This is usually a human-readable name of the Keycloak client.
    type: str
    required: true
  realm:
    description:
      - The name of the Keycloak realm the Keycloak client is in.
    type: str
    required: true

extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Samuli Seppänen (@mattock)
"""

EXAMPLES = r"""
- name: Manage scope-based Keycloak authorization permission
  community.general.keycloak_authz_permission:
    name: ScopePermission
    state: present
    description: Scope permission
    permission_type: scope
    scopes:
      - file:delete
    policies:
      - Default Policy
    client_id: myclient
    realm: myrealm
    auth_keycloak_url: http://localhost:8080/auth
    auth_username: keycloak
    auth_password: keycloak
    auth_realm: master

- name: Manage resource-based Keycloak authorization permission
  community.general.keycloak_authz_permission:
    name: ResourcePermission
    state: present
    description: Resource permission
    permission_type: resource
    resources:
      - Default Resource
    policies:
      - Default Policy
    client_id: myclient
    realm: myrealm
    auth_keycloak_url: http://localhost:8080/auth
    auth_username: keycloak
    auth_password: keycloak
    auth_realm: master
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str

end_state:
  description: Representation of the authorization permission after module execution.
  returned: on success
  type: dict
  contains:
    id:
      description: ID of the authorization permission.
      type: str
      returned: when O(state=present)
      sample: 9da05cd2-b273-4354-bbd8-0c133918a454
    name:
      description: Name of the authorization permission.
      type: str
      returned: when O(state=present)
      sample: ResourcePermission
    description:
      description: Description of the authorization permission.
      type: str
      returned: when O(state=present)
      sample: Resource Permission
    type:
      description: Type of the authorization permission.
      type: str
      returned: when O(state=present)
      sample: resource
    decisionStrategy:
      description: The decision strategy to use.
      type: str
      returned: when O(state=present)
      sample: UNANIMOUS
    logic:
      description: The logic used for the permission (part of the payload, but has a fixed value).
      type: str
      returned: when O(state=present)
      sample: POSITIVE
    resources:
      description: IDs of resources attached to this permission.
      type: list
      returned: when O(state=present)
      sample:
        - 49e052ff-100d-4b79-a9dd-52669ed3c11d
    scopes:
      description: IDs of scopes attached to this permission.
      type: list
      returned: when O(state=present)
      sample:
        - 9da05cd2-b273-4354-bbd8-0c133918a454
    policies:
      description: IDs of policies attached to this permission.
      type: list
      returned: when O(state=present)
      sample:
        - 9da05cd2-b273-4354-bbd8-0c133918a454
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    get_token,
    keycloak_argument_spec,
)


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    meta_args = dict(
        state=dict(type="str", default="present", choices=["present", "absent"]),
        name=dict(type="str", required=True),
        description=dict(type="str"),
        permission_type=dict(type="str", choices=["scope", "resource"], required=True),
        decision_strategy=dict(type="str", default="UNANIMOUS", choices=["UNANIMOUS", "AFFIRMATIVE", "CONSENSUS"]),
        resources=dict(type="list", elements="str", default=[]),
        scopes=dict(type="list", elements="str", default=[]),
        policies=dict(type="list", elements="str", default=[]),
        client_id=dict(type="str", required=True),
        realm=dict(type="str", required=True),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"]]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    # Convenience variables
    state = module.params.get("state")
    name = module.params.get("name")
    description = module.params.get("description")
    permission_type = module.params.get("permission_type")
    decision_strategy = module.params.get("decision_strategy")
    realm = module.params.get("realm")
    client_id = module.params.get("client_id")
    realm = module.params.get("realm")
    resources = module.params.get("resources")
    scopes = module.params.get("scopes")
    policies = module.params.get("policies")

    if permission_type == "scope" and state == "present":
        if scopes == []:
            module.fail_json(msg="Scopes need to defined when permission type is set to scope!")
        if len(resources) > 1:
            module.fail_json(msg="Only one resource can be defined for a scope permission!")

    if permission_type == "resource" and state == "present":
        if resources == []:
            module.fail_json(msg="A resource need to defined when permission type is set to resource!")
        if scopes != []:
            module.fail_json(msg="Scopes cannot be defined when permission type is set to resource!")

    result = dict(changed=False, msg="", end_state={})

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    # Get id of the client based on client_id
    cid = kc.get_client_id(client_id, realm=realm)
    if not cid:
        module.fail_json(msg=f"Invalid client {client_id} for realm {realm}")

    # Get current state of the permission using its name as the search
    # filter. This returns False if it is not found.
    permission = kc.get_authz_permission_by_name(name=name, client_id=cid, realm=realm)

    # Generate a JSON payload for Keycloak Admin API. This is needed for
    # "create" and "update" operations.
    payload = {}
    payload["name"] = name
    payload["description"] = description
    payload["type"] = permission_type
    payload["decisionStrategy"] = decision_strategy
    payload["logic"] = "POSITIVE"
    payload["scopes"] = []
    payload["resources"] = []
    payload["policies"] = []

    if permission_type == "scope":
        # Add the resource id, if any, to the payload. While the data type is a
        # list, it is only possible to have one entry in it based on what Keycloak
        # Admin Console does.
        r = False
        resource_scopes = []

        if resources:
            r = kc.get_authz_resource_by_name(resources[0], cid, realm)
            if not r:
                module.fail_json(
                    msg=f"Unable to find authorization resource with name {resources[0]} for client {cid} in realm {realm}"
                )
            else:
                payload["resources"].append(r["_id"])

            for rs in r["scopes"]:
                resource_scopes.append(rs["id"])

        # Generate a list of scope ids based on scope names. Fail if the
        # defined resource does not include all those scopes.
        for scope in scopes:
            s = kc.get_authz_authorization_scope_by_name(scope, cid, realm)
            if r and s["id"] not in resource_scopes:
                module.fail_json(
                    msg=f"Resource {resources[0]} does not include scope {scope} for client {client_id} in realm {realm}"
                )
            else:
                payload["scopes"].append(s["id"])

    elif permission_type == "resource":
        if resources:
            for resource in resources:
                r = kc.get_authz_resource_by_name(resource, cid, realm)
                if not r:
                    module.fail_json(
                        msg=f"Unable to find authorization resource with name {resource} for client {cid} in realm {realm}"
                    )
                else:
                    payload["resources"].append(r["_id"])

    # Add policy ids, if any, to the payload.
    if policies:
        for policy in policies:
            p = kc.get_authz_policy_by_name(policy, cid, realm)

            if p:
                payload["policies"].append(p["id"])
            else:
                module.fail_json(
                    msg=f"Unable to find authorization policy with name {policy} for client {client_id} in realm {realm}"
                )

    # Add "id" to payload for update operations
    if permission:
        payload["id"] = permission["id"]

        # Handle the special case where the user attempts to change an already
        # existing permission's type - something that can't be done without a
        # full delete -> (re)create cycle.
        if permission["type"] != payload["type"]:
            module.fail_json(
                msg=(
                    f"Modifying the type of permission (scope/resource) is not supported: "
                    f"permission {permission['id']} of client {cid} in realm {realm} unchanged"
                )
            )

    # Updating an authorization  permission is tricky for several reasons.
    # Firstly, the current permission is retrieved using a _policy_ endpoint,
    # not from a permission endpoint. Also, the data that is returned is in a
    # different format than what is expected by the payload. So, comparing the
    # current state attribute by attribute to the payload is not possible.  For
    # example the data contains a JSON object "config" which may contain the
    # authorization type, but which is no required in the payload.  Moreover,
    # information about resources, scopes and policies is _not_ present in the
    # data. So, there is no way to determine if any of those fields have
    # changed. Therefore the best options we have are
    #
    # a) Always apply the payload without checking the current state
    # b) Refuse to make any changes to any settings (only support create and delete)
    #
    # The approach taken here is a).
    #
    if permission and state == "present":
        if module.check_mode:
            result["msg"] = "Notice: unable to check current resources, scopes and policies for permission. \
                            Would apply desired state without checking the current state."
        else:
            kc.update_authz_permission(
                payload=payload, permission_type=permission_type, id=permission["id"], client_id=cid, realm=realm
            )
            result["msg"] = "Notice: unable to check current resources, scopes and policies for permission. \
                            Applying desired state without checking the current state."

        # Assume that something changed, although we don't know if that is the case.
        result["changed"] = True
        result["end_state"] = payload
    elif not permission and state == "present":
        if module.check_mode:
            result["msg"] = "Would create permission"
        else:
            kc.create_authz_permission(payload=payload, permission_type=permission_type, client_id=cid, realm=realm)
            result["msg"] = "Permission created"

        result["changed"] = True
        result["end_state"] = payload
    elif permission and state == "absent":
        if module.check_mode:
            result["msg"] = "Would remove permission"
        else:
            kc.remove_authz_permission(id=permission["id"], client_id=cid, realm=realm)
            result["msg"] = "Permission removed"

        result["changed"] = True

    elif not permission and state == "absent":
        result["changed"] = False
    else:
        module.fail_json(
            msg=f"Unable to determine what to do with permission {name} of client {client_id} in realm {realm}"
        )

    module.exit_json(**result)


if __name__ == "__main__":
    main()
